 -------------------------- MODULE Actions ---------------------------
\* This module defines all the actions

(* The hierarchy of the modules are
 Ring (The basic parameters and operations for ring calculation)
   \subset LS (The leaf set utilities)
      \subset RT (The routing table utilities)
         \subset Msg (The messages definitions)
            \subset InitialStates (The initial state and TypeInvariant)
               \subset Actions (All the actions)
                  \subset MSPastry (The specification of MSPastry including properties)
                     \subset MCMSPastry (The demonstration assignments for MSPastry)
*)
EXTENDS InitialStates
---------------------------------------------------------------
\* Application relevant actions

\* an application on node i tries to find the node k 
\* or to find the node which is closest to the key k.
\* The node i generate a Lookup msg and ask himself to
\* route it to k. 
Lookup(i, k) ==
  /\ receivedMsgs'= receivedMsgs \cup 
                    {[destination |-> i, 
                      mreq |-> [
                                type |-> "Lookup", 
                                node |-> k]
                     ]}
  /\ UNCHANGED <<active, rtable, lset, probing, failed>>


\* a node i has got an ID "seed" from the application running based
\* on the overlay network and he wants to join the network.
\* Seed is one of the active nodes in the network who is in charge of
\* some range of keys which is numerrically close to him.
\* The node i generate the JoinRequest msg for seed to let him
\* route the msg to the ID i. By doing this, the msg will be replied
\* by the node nummerically closest to i and hence find the correct
\* place for i to join into the network.

CastroJoin(i, seed) == 
  /\ i # seed  
  /\ active[i] = FALSE /\ active[seed] =TRUE          
  /\ receivedMsgs'= receivedMsgs \cup 
                    {[destination |-> seed, 
                      mreq |-> [
                                type |-> "JoinRequest", 
                               rtable |-> InitRTable, 
                               node |-> i]
                     ]}
  /\ UNCHANGED <<rtable, lset, probing, failed, active>>  
------------------------------------------------
\* Internal behavior of system 

\* (1) Delivering 
\* a node i deliver the lookup message it take charge of k
\* Only the active node can deliver Lookup messages,
\* which destination is leaseed by this node
CastroDeliver(i, k) == 
  /\ active[i] = TRUE
  /\ \E m \in receivedMsgs: 
     /\ m.mreq.type = "Lookup"
     /\ m.destination = i
     /\ m.mreq.node = k
     /\ Covers(lset[i], k)
     /\ receivedMsgs'= (receivedMsgs \{m}) 
  /\ UNCHANGED <<active, rtable, lset, probing, failed>>  
---------------------------------------------------------
\* (2) Routing

\*Help function for route
FindNext(i, j) ==
\* simplified version
  LET 
     lsCan ==  GetLSetContent(lset[i]) \ failed[i]
     canrelax ==
           {can \in (GetLSetContent(lset[i]) 
                    \cup GetRTableContent(rtable[i]))
                    \ failed[i]:
                       AbsDist(j, can) < AbsDist(j, i)}
  IN 
   
    IF \* We should find the next hop in leaf set
     \* when the leaf set of i covers whole ring,
     \* or i is within the range of leaf set;
      \/ Overlaps(lset[i])      
      \/ CwDist(LeftMost(lset[i]), j) =< CwDist(LeftMost(lset[i]), RightMost(lset[i]))
     \*/\ lsCan # {} \* this set cannot be empty because it contains i which cannot be 
                   \* inside failed[i]
    THEN CHOOSE n \in lsCan: \A m \in lsCan: AbsDist(n, j) =< AbsDist(m, j)
  \* Real network will check the liveness via underlying TCP PING, but here
  \* we assume the TCP is reliable and we pretend to get the liveness
    ELSE \*Otherwise we should try the rtable
       IF canrelax # {} 
       THEN CHOOSE can \in canrelax: \A m \in canrelax: AbsDist(can, j) =< AbsDist(m, j)
       ELSE i
\* The following is the complex algorithm involves routing table lookup
(* To get the next forward destination.*)
\* since we check the liveness of a node before pick it as candidate, 
\* we don't take it if it is FALSE
(*
  IF \* We should find the next hop in leaf set
     \* when the leaf set of i covers whole ring,
     \* or i is within the range of leaf set;
     \/ Overlaps(lset[i])      
     \/ CwDist(LeftMost(lset[i]), j) =< CwDist(LeftMost(lset[i]), RightMost(lset[i]))
  THEN LET lsCan == GetLSetContent(lset[i]) \ failed[i] \* if leafset of i is empty then i \in lsCan
   \*{c \in GetLSetContent(lset[i]) \ failed[i]: status[c] # "dead"}
  \* Real network will check the liveness via underlying TCP PING, but here
  \* we assume the TCP is reliable and we pretend to get the liveness
       IN CHOOSE n \in lsCan: \A m \in lsCan: AbsDist(n, j) =< AbsDist(m, j)
  ELSE \*Otherwise we should try the rtable
       LET r == SharedDig(j, i)
           rtCan == rtable[i][<<r, GetDig(j, r+1)>>]
       IN IF rtCan # null \*/\ status[rtCan] # "dead"
          THEN rtCan
          ELSE \* If we cannot find anyone from rtable, we relax the condition and try with 
               \* all we know from rtable and lset
               LET canrelax ==
                   {can \in (GetLSetContent(lset[i]) 
                            \cup GetRTableContent(rtable[i]))
                            \ failed[i]:
                   AbsDist(j, can) < AbsDist(j, i) /\ SharedDig(j, can) >= r}\* /\ status[can] # "dead"}
                   \* again, we simulate the check liveness here
               IN IF canrelax # {} 
                  THEN CHOOSE can \in canrelax: \A m \in canrelax: AbsDist(can, j) =< AbsDist(m, j)
                  ELSE i
*)

\* the node i forwards a look up to a closer node to the final destination k
\* when it is TRUE and the destination key is outside of its range
\* If we can't find the closer node, then we fire an Error
RouteLookup(i, k) == 
  /\ \E m \in receivedMsgs:
     /\ m.destination = i
     /\ m.mreq.type ="Lookup" 
     /\ m.mreq.node = k 
     /\ ~Covers(lset[i], k)
     \* the k is not in the cover range of i
     /\ LET nh ==  FindNext(i, k)
        IN  receivedMsgs'= (receivedMsgs \{m}) \cup 
              IF nh # i /\ nh # null
              THEN {[destination |-> nh, mreq |-> m.mreq]}
              ELSE {[destination |-> i, mreq |-> [type |-> "NoLegalRoute", key |-> k]]}\*fire the error
  /\ UNCHANGED <<active, rtable, lset, probing, failed>>
          
\* the node i forwards a join request to a closer node to the final destination k
\* when it is TRUE and the destination key is outside of its range
\* If we can't find the closer node, then we fire an Error
RouteJReq(i, j) == 
  \*/\ active[i] = TRUE
  /\ \E m \in receivedMsgs:
     /\ m.mreq.type = "JoinRequest"
     /\ m.destination = i
     /\ m.mreq.node = j 
     \* /\ ~k \in GetLSetContent(lset[i]) 
        \* to avoid the case that a node left and rejoin 
        \* before its neighbor deleted it from their leaf set
        \* but this should not cause a real problem. If I block the join, then the rejoined node will never be able to join.
     /\ ~Covers(lset[i], j)
     /\ LET nh ==  FindNext(i, j)
        IN 
        receivedMsgs'= (receivedMsgs \{m}) \cup 
          IF nh # i
          THEN {[destination |-> nh, mreq |-> 
                [type |-> "JoinRequest",
               rtable |-> AddToTable(
          GetRTableContent(rtable[i]), m.mreq.rtable, i), node |-> j]]}
               \*for join request, we update the rtable by each hop
          ELSE {[destination |-> i, mreq |-> 
                [type |-> "NoLegalRoute", key |-> j]]}
  /\ UNCHANGED <<active, rtable, lset, probing, failed>>                
-----------------------------------------------------------------------------------------------
\* (3) Consistent Join

\* node i receives join request msg and send back join reply with its leaf set
\* when the node i is TRUE and the destination is within the cover range
CastroRecJReq(i) ==  
  /\ active[i] = TRUE 
  /\ \E m \in receivedMsgs:
   /\ m.mreq.type = "JoinRequest"
   /\ m.destination = i
   /\ Covers(lset[i], m.mreq.node)
   /\ LET newmjr == [type |-> "JoinReply",  
                            rtable |-> m.mreq.rtable,
                              lset |-> lset[i]]
          newmj == [destination |-> m.mreq.node,
                           mreq |-> newmjr]
      IN receivedMsgs'= (receivedMsgs \{m}) \cup {newmj}
  /\ UNCHANGED <<active, rtable, lset, probing, failed>>
   
\* A function to build the Probe Message to its neighbour
Probe(i, ls, f, toprob) == 
  {[destination |-> j,
          mreq |-> [type |-> "LSProbe",  
                    node |-> i,
                    lset |-> ls,
                  failed |-> f
                   ]
  ]: j\in toprob}
  
\* Node i receives join reply msg, it updates its rtable and lset with
\* the received information and probe the new nodes in its leaf set.
RecJReply(i) == 
\* 1) A node shouldnot receive join reply if it is no more in the network (FALSE)
\*    to avoid to be "activated" by gabage message.
\* 2) To receive join reply after it turns to be "ok" or TRUE makes no sense.
/\ \E m \in receivedMsgs: 
   /\ m.destination = i
   /\ m.mreq.type = "JoinReply"
   /\ LET
        newrtable ==  AddToTable(GetLSetContent(m.mreq.lset)
                                \cup GetRTableContent(m.mreq.rtable), 
                                 rtable[i], i)
        newlset == AddToLSet(GetLSetContent(m.mreq.lset), lset[i])
        toprob == GetLSetContent(newlset) \ {i}
        probmsgs == Probe(i, newlset, {}, toprob)
      IN
     /\ rtable'= [rtable EXCEPT ![i] = newrtable] 
     /\ lset'= [lset EXCEPT ![i] = newlset] 
     /\ probing'= [probing EXCEPT ![i] = toprob]
     /\ failed'=[failed EXCEPT ![i] = {}]
     /\ receivedMsgs'= (receivedMsgs \ {m}) \cup probmsgs
/\ UNCHANGED <<active>>
     
\* Node i receives probe msg from j. 
\* It updates the local knowledge w.r.t. j and send reply back to j.
\* It probes nodes in leaf set which are in failed and removes them from its leaf set.(prb1)
\* Then it creates the clone L' of its leaf set and add nodes in L, which it 
\* consideres not to be faulty. (lprim)
\* It probes the new nodes in lprim before inclusion. (prb2)
RecProbe(i) ==
/\ \E m \in receivedMsgs:
  /\ m.destination = i
  /\ m.mreq.type = "LSProbe"
  /\ LET j == m.mreq.node
        fi == failed[i] \ {j} \*failed[i] removes j
       ls1 == AddToLSet({j}, lset[i]) \* lset[i] extended with j
       ls2 == RemoveFromLSet(m.mreq.failed, ls1)\*lset[i] removes the common members of m.failed
     lprim == GetLSetContent(AddToLSet((GetLSetContent(m.mreq.lset) \ fi), ls2))\* see above
      prb1 == (GetLSetContent(ls1) \cap m.mreq.failed )\ (probing[i] \cup fi) 
      prb2 == lprim \ (GetLSetContent(ls2) \cup probing[i] \cup fi \cup prb1)
      newm == [type |-> "LSProbeReply",  
                  node |-> i,
                  lset |-> ls2,
                  failed |-> fi]
     IN
         /\failed' = [failed EXCEPT ![i]=fi]
         /\rtable' = [rtable EXCEPT ![i] = AddToTable({j}, @, i)]
         /\lset'   = [lset EXCEPT ![i] = ls2]
         /\probing'= [probing EXCEPT ![i] = (@ \cup prb1  \cup prb2) \ fi] 
         \* If node i has probed j and waiting its reply but then get the probe msg from j
         \* it continue to wait j's reply. So, the probing' might contains j          
         \* In case we removed the nodes of m.mreq.failed from leaf set of i, then it is no more complete
         \* we need to change the node's active in order to complete the leaf set first before it can route badly
         \* see vE-pastry/log/log-receiveProbeWithFailedNodes.txt since step 19, when node 17 received msg from 18
         \* containing the failed node 95, node 17 removes 95 from its leaf set, then it is no more completed
         /\receivedMsgs' = (receivedMsgs \ {m}) 
                           \cup { [destination |-> j, mreq |-> newm]} 
                           \cup Probe(i, ls1, fi, prb1)
                           \cup Probe(i, ls2, fi, prb2)
         /\ UNCHANGED active
\* A node i receives probe reply msg. It does similar things as receiving probe msg except:
\* instead of sending reply, it removes j from probing; (prb3)
\* If there is still out standing probes, do nothing. (prb3 = {})
\* If no out standing probes exists, it checks if its leaf set is complete
\* when yes, it turns its active to "ok" and thus finish the consistent join phrase.
\* If the leaf set is not complete, it probes the farest nodes (leftmost and rightmost) 
\* in leaf set to complete it.
RecPRplyInner(i,m) ==
  LET j == m.mreq.node
      Ls == m.mreq.lset
      fi == failed[i] \ {j} \*failed[i] removes j
     ls1 == AddToLSet({j}, lset[i]) \* lset[i] extended with j
     ls2 == RemoveFromLSet(m.mreq.failed, ls1)\*lset[i] removes the common members of m.failed
   lprim == GetLSetContent(AddToLSet((GetLSetContent(Ls) \ fi), ls2))
    prb1 == (GetLSetContent(ls1) \cap m.mreq.failed )\ (probing[i] \cup fi)
    prb2 == lprim \ (GetLSetContent(ls2) \cup probing[i] \cup fi \cup prb1)
    prb3 == ((probing[i] \cup prb1 \cup prb2) \ failed[i]) \ {j} 
    cand == (GetLSetContent(ls2) \cup GetRTableContent(rtable[i])) \ {i}
      lm == {LeftMost(lset[i])}
      rm == {RightMost(lset[i])}
 newprob == IF Overlaps(ls2) \/ IsComplete(ls2)
            THEN {}
            ELSE IF Lenth(ls2.left) < L /\ Lenth(ls2.right) < L
                 THEN lm \cup rm
                 ELSE IF Lenth(ls2.left) < L /\ Lenth(ls2.right) = L
                      THEN lm
                      ELSE rm
shouldBeOK == \* A node has no outstanding probe and has completed its leaf set
              \* then it should turn to be active ("ok")
              /\ active[i] = FALSE
              /\ prb3={}
              /\ ( Overlaps(ls2) \/ IsComplete(ls2))
  IN /\ rtable' = [rtable EXCEPT ![i] = AddToTable({j}, @, i)]
     /\ lset'   = [lset EXCEPT ![i] = ls2]
     /\ failed' = [failed EXCEPT ![i] = IF prb3={} /\ IsComplete(ls2) THEN {} ELSE fi]       
     /\ probing'= [probing EXCEPT ![i] = IF ~prb3={} THEN prb3 ELSE newprob]
     /\ active'= [active EXCEPT ![i] = IF shouldBeOK THEN TRUE ELSE @]
     /\ receivedMsgs' = (receivedMsgs \ {m}) 
                        \cup Probe(i, ls1, fi, prb1)
                        \cup Probe(i, ls2, fi, prb2)
                        \cup Probe(i, ls2, fi, newprob)

RecPRply(i) ==
\* only FALSE nodes donnot react on probe reply.
/\ \E m \in receivedMsgs: 
   /\ m.destination = i
   /\ m.mreq.type = "LSProbeReply" 
   /\ m.mreq.node \in probing[i] \* others' faked probe-reply might make i active
                                 \* and hence create a partition of network
   /\ RecPRplyInner(i,m)
       
\* A node i suspected j to be faulty and hence probe j
SuspectFaulty(i, j) == 
  /\ active[i] # FALSE \* only non-faulty node can suspect
  /\ active[j]= FALSE \* For simplicity we enforce it. 
  /\ i # j \* node doesnot suspect itself
  /\ j \in GetLSetContent(lset[i]) \* node only suspect nodes in its leaf set
  /\ j \notin probing[i] \*node does not suspect outstanding probed node
  /\ j \notin failed[i] \*node does not suspect a 'dead' node it knows
  /\ receivedMsgs'= receivedMsgs \cup Probe(i, lset[i], failed[i], {j})
  /\ probing'=[probing EXCEPT ![i]= @\cup {j}]
  /\ UNCHANGED <<rtable, lset, failed, active>>

\* Simulation of a normal behavior.
\* A node i stops waiting reply from node j and considers j to be failed  
\* In the real system, 
\* the node i will try several times before it gives up waiting. 
\* Besides it will try different ways
\* to contact node j, including using virtual link techniques.
\* We simplified the virtual link technique in our model:
\* If a node has blocked incoming channel but not outgoing channel,
\* Then we assume that a virtual link will be set up so that it behaves
\* exactly like a normal node. Hence, after using virtual link, 
\* the node can contact each other. Hence all the probe will be ordinarily handled.
\* In all, as soon as probe time out happend by i for probing j, it means
\* there is no way for i to find j, even through virtual link. 
\* Then there is only two possibilities: either i has already left the
\* network or j has. 
\* The first case is ruled out by previous gard, so j must have left the network.
\* That's why we simulate the real situation like this. 
TimeOut(i, j)== 
/\ active[i]# FALSE \* for nodes no more in the network, it does not matter
/\ active[j]= FALSE \* For simplicity we enforce it. 
/\ j \in probing[i]
/\ failed' = [failed EXCEPT ![i] = @\cup {j}]
/\ probing' = [probing EXCEPT ![i] = @ \ {j}]
/\ lset' = [lset EXCEPT ![i] = RemoveFromLSet({j},@) ]
/\ receivedMsgs' = receivedMsgs \
                          IF \E m \in receivedMsgs: m.mreq.type = "LSProbe"
\* here we delete all the gabage probe msgs from i to j, since 
\* we are sure no one will get a reply any more. 
                                /\ m.mreq.node = i
                                /\ m.destination = j
                          THEN  {m \in receivedMsgs: m.mreq.type = "LSProbe"
                                   /\ m.mreq.node = i
                                   /\ m.destination = j}
       
                          ELSE {}
/\UNCHANGED <<rtable, active>>

LSRepair(i) == 
\* find the new candidate for the missing entry in leaf set and try to complete it
/\ active[i] # FALSE
/\ probing[i] = {}
/\ ~IsComplete(lset[i])
/\ Lenth(lset[i].left) # 0  /\ Lenth(lset[i].right) # 0 
/\ LET
   lm == {LeftMost(lset[i])}
   rm == {RightMost(lset[i])}
   newprob == IF Lenth(lset[i].left) < L /\ Lenth(lset[i].right) < L 
              THEN lm \cup rm
              ELSE IF Lenth(lset[i].left) < L /\ Lenth(lset[i].right) = L
                   THEN lm
                   ELSE rm \* here it contains the case when leaf set are complete 
                           \* because of overlap happens
   IN 
   /\ active'= [active EXCEPT ![i] = FALSE] 
   /\ probing' = [probing EXCEPT ![i] =  probing[i] \cup newprob]
   /\ receivedMsgs' = receivedMsgs \cup Probe(i, lset[i], failed[i], newprob)
/\ UNCHANGED <<lset, rtable, failed>>
----------------------------------------------------------------
\* (5) Failure simulation actions
MsgLost == 
/\\E m \in receivedMsgs:
   \*/\ PrintT(<<"MsgLost", m>>)
   /\ receivedMsgs' = receivedMsgs \ {m}
/\ UNCHANGED <<active, rtable, lset, probing, failed>>
NodeLeft(i) ==
  /\ active[i] = TRUE
  \*/\ active[i] # FALSE \* a dead node has already left
  /\ \E j \in I: j # i /\ active[j] = TRUE\*to make sure there is at lease one node working
  /\ active' = [active EXCEPT ![i] = FALSE]
  /\ rtable' = [rtable EXCEPT ![i] = AddToTable({i}, InitRTable, i)]
  /\ lset' = [lset EXCEPT ![i] = EmptyLS(i)]
  /\ probing' = [probing EXCEPT ![i] = {}]
  /\ failed' = [failed EXCEPT ![i] = {}]
  /\ UNCHANGED <<receivedMsgs>> \* messages sent to it should not be removed, 
  \* it is buffered somewhere in the network, and will be received by the left node
  \* whenever it becomes active again.
==============================================================